#include <LiquidCrystal.h>

//constants and global variables
  //LCD
    const int rs = 12, en = 7, d4 = 5, d5 = 4, d6 = 8, d7 = 2;
    const int lcd_red=3, lcd_green=11, lcd_blue=6;
    const int n_r=110, n_g=125, n_b=125;
    const int o_r=0, o_g=255, o_b=0;
    const int c_r=255, c_g=123, c_b=0;
    const int w_r=255, w_g=0, w_b=0;
    enum farba {neutral=0, ok=1, caution=2, warning=3};
    farba current_color=caution;
    LiquidCrystal lcd(rs, en, d4, d5, d6, d7);
  //MQ7
    const int MQ7_output = A1, MQ7_control=9;
    int CO_value=0;
    float MQ7_change=0;
    enum MQ7_status {heating, cooling};
    MQ7_status actual_status=heating;
    const int border_ok=255, border_caution=270;
  //IR and RED measurment
    int T=20, sensorPin=A0, IR_LED=10, RED_LED=13;
    const int max_measurments_size=50, min_inserted_finger=200;
    float RED_measurments[max_measurments_size], IR_measurments[max_measurments_size];
    int index=0, measurments_size=0;
  //BPM
    float sum_bpms=0, last_peak=0;
    const int max_bpms=5, avg_const=4, diff=0.21, min_diff_between_peak=400;
    const int dif_prev=1, dif_prevprev=2, dif_last=4;
    int bpms=0, index_bpm=0;
    float bpm[max_bpms];
    float BPM_value=0;
  //R and SPO2
    float REDmin=30000, REDmax=0, IRmin=30000, IRmax=0, R=0, sum_Rs=0;
    float Rs[max_bpms];
    float k=-22, q=105;
    float SPO2_value=0;
    const int border_ok_spo2=92, border_caution_spo2=88;
//end constants and global variables

void setup() {
  Serial.begin(9600);
  Serial.flush();
  Serial.println("beat, R, SPO2, BPM"); 
    
  pinMode(sensorPin,INPUT);
  pinMode(RED_LED,OUTPUT);
  pinMode(IR_LED,OUTPUT);
  
  pinMode(MQ7_control, OUTPUT);

  pinMode(en, OUTPUT);
  pinMode(lcd_red, OUTPUT);
  pinMode(lcd_green, OUTPUT);
  pinMode(lcd_blue, OUTPUT);

  lcd_backlight(neutral);
  
  digitalWrite(RED_LED,LOW);
  digitalWrite(IR_LED,LOW);
  analogWrite(MQ7_control, 255);
  MQ7_change=millis();
  actual_status= heating;

  lcd.begin(16, 2);
  
  lcd.setCursor(0,1); 
  lcd.print("Starting");
  delay(500);
  lcd.clear();
  
  for(int i=0;i<max_bpms;i++)
    bpm[i]=Rs[i]=0;
}

void loop() {
  current_color=neutral;
  float IR = measure(IR_LED, sensorPin);
  Serial.print(",");
  Serial.print(IR);
  float RED = measure(RED_LED, sensorPin);
  Serial.print(",");
  Serial.print(RED);
  if(IR>min_inserted_finger && RED>min_inserted_finger)
  {
    write_measurment(IR, RED);
  
    change_max_min_for_R(RED, IR);
    if(measurments_size==max_measurments_size)
      compute_blood_stuff();
  }

  MQ7_stuff();

  lcd_backlight(current_color);
  lcd_write();
  
  Serial.println();
}

void lcd_write()
{
  lcd.setCursor(0,0);
  lcd.print("   ");
  if(BPM_value<100)
    lcd.setCursor(1,0);
  else
    lcd.setCursor(0,0);
  lcd.print(BPM_value);
  lcd.setCursor(3,0);
  lcd.print("BPM");
  
  lcd.setCursor(10,0);
  lcd.print("   ");
  if(SPO2_value<100)
    lcd.setCursor(11,0);
  else
    lcd.setCursor(10,0);
  lcd.print(SPO2_value);
  lcd.setCursor(12,0);
  lcd.print("SPO2");
  
  lcd.setCursor(0,1);
  lcd.print("   ");
    if(CO_value<100)
    lcd.setCursor(1,1);
  else
    lcd.setCursor(0,1);
  lcd.print(CO_value);
  lcd.setCursor(3,1);
  lcd.print("ppm");
  
  lcd.setCursor(11,1);
  if(actual_status==heating)
    lcd.print("H");
  else
    lcd.print("C");
  
  lcd.print(":");
  
  lcd.print((int)(MQ7_change-millis())/1000);
  lcd.print("s");
}

void MQ7_stuff()
{
  if(millis()>MQ7_change)
  {
    if(actual_status==heating)
    {
      actual_status=cooling;
      analogWrite(MQ7_control, 100);
      MQ7_change=millis()+9000;
    }
    else if(actual_status==cooling)
    {
      actual_status=heating;
      analogWrite(MQ7_control, 255);
      CO_value=analogRead(MQ7_output);
      //Serial.print(",");
      //Serial.print(CO_value);
      MQ7_change=millis()+6000;
    }    
  }
  if(CO_value<border_ok)
    current_color=max(current_color, ok);
  else if(CO_value>border_caution)
    current_color=max(current_color, warning);
  else
    current_color=max(current_color, caution);
}

float measure(int light_port, int input_port)
{
  digitalWrite(light_port, HIGH);
  float start=millis();
  double res=0;
  int n=0;
  while(millis()<start+T)
  {
    res+=analogRead(input_port);
    n++;
  }
  digitalWrite(light_port, LOW);
  return res/n;
}

void write_measurment(float IR, float RED)
{
  RED_measurments[index]=RED;
  IR_measurments[index]=IR;
  index++;
  
  if(measurments_size<max_measurments_size)
    measurments_size++;
  index%=max_measurments_size;
}

bool is_peak()
{
  float now =avg_ago(index);
  float prev=avg_ago(index-dif_prev);
  float prevprev=avg_ago(index-dif_prevprev);
  float last=avg_ago(index-dif_last);
  
  if(last+diff<prevprev && prevprev+diff<prev && prev>now+diff)
    return true;
  return false;
}

float avg_ago(int index)
{
  float res=0;
  for(int i=avg_const-1;i>=0;i--)
  {
    res+=RED_measurments[(index-i+max_measurments_size)%max_measurments_size];
    res+=IR_measurments[(index-i+max_measurments_size)%max_measurments_size];
  }
  return res/(2*avg_const);
}

void compute_blood_stuff()
{
  if(is_peak() && millis()- last_peak>min_diff_between_peak)
  {
    lcd.setCursor(7,0);
    lcd.print("<3");
    Serial.print("\n");
    Serial.print(960);
    sum_bpms-=bpm[index_bpm];
    sum_Rs-=Rs[index_bpm];
    
    sum_bpms+=(bpm[index_bpm]=millis()- last_peak);
    sum_Rs+=(Rs[index_bpm] = (((REDmax-REDmin)*IRmin) / ((IRmax-IRmin)*REDmin)));

    REDmin=30000, REDmax=0, IRmin=30000, IRmax=0;    
    last_peak=millis();
    
    index_bpm++;
    index_bpm%=max_bpms; 
  }
  else
  {
    if(millis()- last_peak>min_diff_between_peak/3*2)
    {
      lcd.setCursor(7,0);
      lcd.print("  ");
    }
    Serial.print("\n");
    Serial.print(920);
  }

  if((k*(sum_Rs/max_bpms)+q)>=border_ok_spo2)
    current_color = ok;
  else if((k*(sum_Rs/max_bpms)+q)<=border_caution_spo2)
    current_color = warning;
  else
    current_color = caution;

  SPO2_value=k*(sum_Rs/max_bpms)+q;
  BPM_value = 60000/(sum_bpms/max_bpms);
  
  /*Serial.print(",");
  Serial.print(sum_Rs/max_bpms);
  Serial.print(",");
  Serial.print(k*(sum_Rs/max_bpms)+q);
  Serial.print(",");
  Serial.print(60000/(sum_bpms/max_bpms));*/
}

void change_max_min_for_R(float RED, float IR)
{
  REDmax=max(RED, REDmax);
  REDmin=min(RED, REDmin);
  IRmax=max(IR, IRmax);
  IRmin=min(IR, IRmin);
}

void lcd_backlight(int r, int g, int b)
{
  analogWrite(lcd_red, r);
  analogWrite(lcd_green, g);
  analogWrite(lcd_blue, b);
}
void lcd_backlight(farba f)
{
  if(f==neutral)
    lcd_backlight(n_r, n_g, n_b);
  if(f==ok)
    lcd_backlight(o_r, o_g, o_b);
  if(f==caution)
    lcd_backlight(c_r, c_g, c_b);
  if(f==warning)
    lcd_backlight(w_r, w_g, w_b);
}
